from typing import Sequence

import cv2
import numexpr as ne
import numpy as np
from onnxruntime import InferenceSession

from ultralight.exceptions import InvalidModelParamException, InvalidInputException
from ultralight.loader import UltraLightLoader
from ultralight.types import SizeType, BoxesType, ScoresType


class UltraLightDetector:

    def __init__(
            self,
            size: SizeType = 640,
            batched: bool = True,
            path: str = None,
            providers: Sequence[str] = None,
            score_threshold: float = 0.9,
            iou_threshold: float = 0.3,
            keep_top_k: int = 1024,
            top_k: int = 256,
    ):
        """
        :param size: size of model input (640x480 or 320x240)
        :param batched: should model use batched input
        :param path: path to ONNX file of model
        :param providers: ONNX providers (devices) as CPUExecutionProvider/CUDAExecutionProvider/CoreMLExecutionProvider
        :param score_threshold: confidence score threshold
        :param iou_threshold: intersection over union threshold
        :param top_k: keep K results before NMS, if k <= 0 than keep all the results
        :param top_k: keep K results after NMS, if k <= 0 than keep all the results
        """
        self._batched = batched
        self._model_shape = (640, 480) if size == 640 else (320, 240)
        self._session = self._create_session(size, batched, path, providers)
        session_input = self._session.get_inputs()[0]
        self._input_name = session_input.name
        self._score_threshold = score_threshold
        self._iou_threshold = iou_threshold
        self._keep_top_k = keep_top_k
        self._top_k = top_k
        onnx_model_shape = tuple(session_input.shape[3:1:-1])
        if onnx_model_shape != self._model_shape:
            raise InvalidModelParamException(
                f'ONNX model\'s shape is {onnx_model_shape}, while detector shape is {self._model_shape}! '
                'Change `size` param or use another model file',
            )
        if self._score_threshold < 0 or self._score_threshold > 1:
            raise InvalidModelParamException(
                f'Invalid score threshold: {self._score_threshold}, it should be in range [0..1]'
            )
        if self._iou_threshold < 0 or self._iou_threshold > 1:
            raise InvalidModelParamException(
                f'Invalid iou threshold: {self._iou_threshold}, it should be in range [0..1]'
            )

    @classmethod
    def _create_session(
            cls,
            size: SizeType,
            batched: bool,
            path: str,
            providers: Sequence[str],
    ) -> InferenceSession:
        if path is None:
            loader = UltraLightLoader()
            path = loader.load(size, batched)
        return InferenceSession(path, providers=providers)

    def detect_one(self, image: np.ndarray, fit_to_image: bool = False) -> tuple[BoxesType, ScoresType]:
        if image is None:
            raise InvalidInputException('Image is None')
        return self.detect_batch(np.asarray([image]), fit_to_image)[0]

    def detect_batch(
            self,
            images: Sequence[np.ndarray],
            fit_to_image: bool = False,
    ) -> list[tuple[BoxesType, ScoresType]]:
        if images is None or len(images) == 0:
            return list()
        for i, image in enumerate(images):
            if image is None:
                raise InvalidInputException(f'Image with index {i} is None')
        if len(images) > 1 and not self._batched:
            raise InvalidInputException(
                f'Current detector is not batched, but images count is more than 1 (count is {len(images)}), '
                f'use `batched=True` param in detector\'s constructor to fix this'
            )
        images_input = self._preprocess_images(images)
        outputs = self._session.run(None, {self._input_name: images_input})
        results = list()
        for image, scores, boxes, *_ in zip(images, *outputs):
            height, width, _ = image.shape
            boxes, scores = self._postprocess(boxes, scores, width, height, fit_to_image)
            results.append((boxes, scores))
        return results

    def _preprocess_images(self, images: Sequence[np.ndarray]) -> np.ndarray:
        images = np.asarray([
            cv2.resize(
                cv2.cvtColor(image, cv2.COLOR_BGR2RGB),
                self._model_shape,
            )
            for image in images
        ])
        images = images.astype(np.float32)  # noqa used in numexpr
        images = ne.evaluate('(images - 127) / 128')
        return np.transpose(images, [0, 3, 1, 2])

    def _postprocess(
            self,
            boxes: np.ndarray,
            scores: np.ndarray,
            width: int,
            height: int,
            fit_to_image: bool = False,
    ) -> tuple[BoxesType, ScoresType]:
        """
        Select boxes that contain human faces (function `predict` in original implementation)
        Parameters:
            boxes (N, 4): boxes array in corner-form
            scores (N, 2): scores array
            width: original image width
            height: original image height
            fit_to_image: should boxes fit to image borders
        Returns:
            faces (K, 4): an array of faces
            scores (K): an array of scores for each face
        """
        scores = scores[:, 1]
        mask = scores > self._score_threshold
        boxes = boxes[mask]
        scores = scores[mask]
        picked = self._hard_nms(boxes, scores)
        if len(picked) == 0:
            return np.asarray([]), np.asarray([], dtype=np.float32)
        boxes = boxes[picked]
        scores = scores[picked]
        boxes *= np.asarray([width, height, width, height])
        if fit_to_image:
            boxes[:, 0::2] = np.clip(boxes[:, 0::2], 0, width - 1)
            boxes[:, 1::2] = np.clip(boxes[:, 1::2], 0, height - 1)
        return boxes.astype(np.int32), scores

    def _hard_nms(self, boxes: np.ndarray, scores: np.ndarray) -> list[int]:
        """
        Perform hard non-maximum-suppression to filter out boxes with iou greater than threshold
        Parameters:
            boxes (N, 4): boxes array in corner-form
            scores (N, 2): scores array
        Returns:
            picked: a list of indexes of the kept boxes
        """
        picked = list()
        indexes = np.argsort(scores)
        indexes = indexes[-self._keep_top_k:]
        while len(indexes) > 0:
            current = indexes[-1]
            picked.append(current)
            if 0 < self._top_k == len(picked) or len(indexes) == 1:
                break
            current_box = boxes[current]
            indexes = indexes[:-1]
            rest_boxes = boxes[indexes]
            iou = self._iou_of(
                rest_boxes,
                np.expand_dims(current_box, axis=0),
            )
            indexes = indexes[iou <= self._iou_threshold]
        return picked

    def _iou_of(self, boxes0: np.ndarray, boxes1: np.ndarray, eps: float = 1e-5) -> np.ndarray:
        """
        Compute intersection-over-union (Jaccard index) of boxes
        Parameters:
            boxes0 (N, 4): ground truth boxes
            boxes1 (N or 1, 4): predicted boxes
            eps: a small number to avoid 0 as denominator
        Returns:
            iou (N): IoU values
        """
        overlap_left_top = np.maximum(boxes0[..., :2], boxes1[..., :2])
        overlap_right_bottom = np.minimum(boxes0[..., 2:], boxes1[..., 2:])
        overlap_area = self._area_of(overlap_left_top, overlap_right_bottom)
        area0 = self._area_of(boxes0[..., :2], boxes0[..., 2:])
        area1 = self._area_of(boxes1[..., :2], boxes1[..., 2:])
        return overlap_area / (area0 + area1 - overlap_area + eps)

    @staticmethod
    def _area_of(left_top: np.ndarray, right_bottom: np.ndarray) -> np.ndarray:
        """
        Compute areas of rectangles by two corners
        Parameters:
            left_top (N, 2): left top corner
            right_bottom (N, 2): right bottom corner
        Returns:
            area (N): return the area
        """
        hw = np.clip(right_bottom - left_top, 0.0, None)
        return hw[..., 0] * hw[..., 1]



import logging
import os.path

import requests
from tqdm import tqdm

from ultralight.types import SizeType

MODEL_FOLDER = '.models'
MODEL_URL_TEMPLATE = 'https://github.com/abionics/UltraLight/releases/download/v2.0.0/{}'


class UltraLightLoader:

    def __init__(self, model_folder: str = MODEL_FOLDER, model_url_template: str = MODEL_URL_TEMPLATE):
        self._model_folder = model_folder
        self._model_url_template = model_url_template

    def load(self, size: SizeType, batched: bool) -> str:
        name = f'ultra_light_{size}_batched.onnx' if batched else f'ultra_light_{size}.onnx'
        filename = os.path.join(self._model_folder, name)
        if not os.path.exists(filename):
            url = self._model_url_template.format(name)
            self._download(url, filename)
        return filename

    def _download(self, url: str, filename: str, block_size: int = 8192):
        os.makedirs(self._model_folder, exist_ok=True)
        logging.info(f'Downloading model "{filename}" from {url}...')
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get('Content-Length', 0))
        with tqdm(total=total_size, unit='iB', unit_scale=True) as pbar:
            pbar.set_description('Downloading model')
            with open(filename, 'wb') as file:
                for data in response.iter_content(block_size):
                    pbar.update(len(data))
                    file.write(data)
        logging.info(f'Downloaded model "{filename}"')



import struct

import cv2
import numpy as np

from ultralight.types import BoxesType, ScoresType


def format_image(image: np.ndarray) -> np.ndarray:
    channels = 1 if len(image.shape) == 2 else image.shape[2]
    if channels == 1:
        return cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    if channels == 4:
        return cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)
    return image


def box_to_point(box: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    return box[0:2], box[2:4]


def draw_faces(image: np.ndarray, boxes: BoxesType, scores: ScoresType = None, **kwargs):
    count = len(boxes)
    show_score = scores is not None
    for i in range(count):
        point_1, point_2 = box_to_point(boxes[i])
        score = scores[i] if show_score else None
        draw_face(image, point_1, point_2, score, **kwargs)


def draw_face(
        image: np.ndarray,
        point_1: np.ndarray,
        point_2: np.ndarray,
        score: float = None,
        score_pad: tuple[int, int] = (0, 5),
        color: tuple[int, int, int] = (0, 0, 255),
        thickness: int = 2,
        font_face: int = cv2.FONT_HERSHEY_SIMPLEX,
        font_scale: float = 0.5,
        **kwargs,
):
    cv2.rectangle(image, point_1, point_2, color, thickness, **kwargs)
    if score is not None:
        position = point_1 - score_pad
        cv2.putText(image, f'{100 * score:.2f}', position, font_face, font_scale, color, thickness, **kwargs)


def create_batched_version(input_filename: str, output_filename: str, batch_size: int | str = 'N'):
    import onnx
    model = onnx.load(input_filename)
    # Set dynamic batch size in inputs and outputs
    graph = model.graph
    for tensor in list(graph.input) + list(graph.value_info) + list(graph.output):
        tensor_dim_0 = tensor.type.tensor_type.shape.dim[0]
        if isinstance(batch_size, int):
            tensor_dim_0.dim_value = batch_size
        else:
            tensor_dim_0.dim_param = batch_size

    # Set dynamic batch size in reshapes (-1)
    for node in graph.node:
        if node.op_type != 'Reshape':
            continue
        for init in graph.initializer:
            # node.input[1] is expected to be a reshape
            if init.name != node.input[1]:
                continue
            # Shape is stored as a list of ints
            if len(init.int64_data) > 0:
                # This overwrites bias nodes' reshape shape but should be fine
                init.int64_data[0] = -1
            # Shape is stored as bytes
            elif len(init.raw_data) > 0:
                shape = bytearray(init.raw_data)
                struct.pack_into('q', shape, 0, -1)
                init.raw_data = bytes(shape)

    # Fix fucking bug (todo fix normally)
    for node in model.graph.node:
        for output in node.output:
            if output not in ('310', '311'):
                continue
            model.graph.output.extend([onnx.ValueInfoProto(name=output)])
    onnx.save(model, output_filename)


if __name__ == '__main__':
    # create_batched_version('../models/ultra_light_320.onnx', '../models/ultra_light_320_batched_64.onnx', 64)
    create_batched_version('../models/ultra_light_640.onnx', '../models/ultra_light_640_batched_4.onnx', 4)



import re

from setuptools import setup, find_packages

REPO_NAME = 'UltraLight'
PACKAGE_NAME = REPO_NAME.lower()
URL = f'https://github.com/abionics/{REPO_NAME}'


def get_version() -> str:
    code = read_file(f'{PACKAGE_NAME}/__init__.py')
    return re.search(r'__version__ = \'(.+?)\'', code).group(1)


def load_readme() -> str:
    return read_file('README.md')


def read_file(filename: str) -> str:
    with open(filename) as file:
        return file.read()


setup(
    name=PACKAGE_NAME,
    version=get_version(),
    description='Ultra Light Fast Generic Face Detector 👨‍👩‍👧‍👦🖼',
    long_description=load_readme(),
    long_description_content_type='text/markdown',
    author='Alex Ermolaev',
    author_email='abionics.dev@gmail.com',
    url=URL,
    license='MIT',
    keywords='face detection ai ultra light',
    install_requires=[
        'numpy',
        'numexpr',
        'opencv-python',
        'requests',
        'tqdm',
        'onnxruntime',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Typing :: Typed',
    ],
    packages=find_packages(exclude=['models', 'samples']),
    zip_safe=False,
)
